<?php

namespace ahrefs\AhrefsSeo;

/**
 * Base class for content settings, implement options get/set.
 */
class Ahrefs_Seo_Content_Settings extends Ahrefs_Seo_Content {

	const OPTION_SCOPE_UPDATED             = 'ahrefs-seo-content-scope-updated';
	const OPTION_SCOPE_LAST_HASH           = 'ahrefs-seo-content-scope-last-hash';
	const OPTION_SCOPE_EXISTING_POST_TYPES = 'ahrefs-seo-content-scope-existing-post-types';
	/**
	 * Set options using global parameters from Wizard form
	 *
	 * @global $_REQUEST
	 * @param bool $return_only Do not save values, just return pages & categories as associative array.
	 * @return null|array null on incorrect request, associative array with pages and categories.
	 */
	public function set_options_from_request( $return_only = false ) {
		$result = [];
        // phpcs:disable WordPress.CSRF.NonceVerification.NoNonceVerification,WordPress.VIP.SuperGlobalInputUsage.AccessDetected,WordPress.Security.NonceVerification.Recommended -- we already checked nonce.
		if ( ! empty( $_REQUEST['ahrefs_audit_options'] ) ) { // assume that some nonce is already checked before this call.
			$waiting_weeks = isset( $_REQUEST['waiting_weeks'] ) ? absint( $_REQUEST['waiting_weeks'] ) : self::DEFAULT_WAITING_WEEKS; // int value.
			$posts = isset( $_REQUEST['post_category'] ) && is_array( $_REQUEST['post_category'] ) ? array_map( 'sanitize_text_field', wp_unslash( $_REQUEST['post_category'] ) ) : []; // int values.
			$products = isset( $_REQUEST['tax_input'] ) && is_array( $_REQUEST['tax_input'] ) && isset( $_REQUEST['tax_input']['product_cat'] ) ? array_map( 'sanitize_text_field', wp_unslash( $_REQUEST['tax_input']['product_cat'] ) ) : []; // int values.
			$pages = isset( $_REQUEST['pages'] ) && is_array( $_REQUEST['pages'] ) ? array_map( 'sanitize_text_field', wp_unslash( $_REQUEST['pages'] ) ) : []; // int values.
			$pages_disabled    = empty( $_REQUEST['pages_enabled'] );
			$posts_disabled    = empty( $_REQUEST['posts_enabled'] );
			$all_cpt_disabled  = empty( $_REQUEST['all_cpt_enabled'] );
			$products_disabled = empty( $_REQUEST['products_enabled'] );
			$cpt_enabled       = isset( $_REQUEST['cpt'] ) && is_array( $_REQUEST['cpt'] ) ? array_map( 'sanitize_text_field', wp_unslash( $_REQUEST['cpt'] ) ) : []; // int values.
            // phpcs:enable WordPress.CSRF.NonceVerification.NoNonceVerification,WordPress.VIP.SuperGlobalInputUsage.AccessDetected,WordPress.Security.NonceVerification.Recommended
			if ( ! $return_only ) {
				$this->set_waiting_weeks( $waiting_weeks );
				$this->set_disabled_for( $pages_disabled, $posts_disabled, $products_disabled, $all_cpt_disabled );
				$this->set_pages_checked( $pages );
				$this->set_posts_categories_checked( $posts );
				$this->set_products_categories_checked( $products );
				$this->set_cpt_checked( $cpt_enabled );
				$hash = (string) md5( (string) wp_json_encode( [ $pages_disabled, $posts_disabled, $products_disabled, $cpt_enabled, $posts, $products, "{$waiting_weeks}" ] ) );
				if ( get_option( self::OPTION_SCOPE_LAST_HASH ) !== $hash ) {
					update_option( self::OPTION_SCOPE_UPDATED, true );
					update_option( self::OPTION_SCOPE_LAST_HASH, $hash );
				}
			}
			if ( ! $pages_disabled ) {
				$result['pages'] = $pages;
			}
			if ( ! $posts_disabled ) {
				$result['categories'] = $posts;
			}
			return $result;
		}
		return null;
	}
	/**
	 * Return not more 11 published pages.
	 *
	 * @return array<int, string> Key is post_id, value is post_title.
	 */
	public function get_pages_list() {
		$result = [];
		/**
		 * @var \WP_Post[] we do not ask for ids.
		 */
		$pages = get_posts(
			[
				'post_status'    => 'publish',
				'post_type'      => 'page',
				'posts_per_page' => 500, // Only first 500 pages.
				'orderby'        => 'title',
				'order'          => 'asc',
			]
		);
		if ( $pages ) {
			foreach ( $pages as $page ) {
				$result[ $page->ID ] = $page->post_title;
			}
		}
		return $result;
	}
	/**
	 * Get registered custom post types, do not include products here.
	 * Return only publicly queryable posts.
	 *
	 * @since 0.8.0
	 *
	 * @param bool $skip_product
	 * @return array<string,string> [slug => label].
	 */
	public function get_custom_post_types( $skip_product = true ) {
		static $result = null;
		if ( is_null( $result ) ) {
			$result = [
				true  => [],
				false => [],
			];
			$cpt    = get_post_types(
				[
					'public'             => true,
					'publicly_queryable' => true,
					'_builtin'           => false,
				],
				'objects',
				'and'
			);
			array_walk(
				$cpt,
				function ( $item ) use ( &$result ) {
					$result[ true ][ $item->name ]  = $item->label;
					$result[ false ][ $item->name ] = $item->label;
				}
			);
			unset( $result[ true ]['product'] );
		}
		return $result[ $skip_product ];
	}
	/**
	 * Turn on options for products if "product" post type exists.
	 *
	 * @since 0.8.0
	 *
	 * @return void
	 */
	public function maybe_turn_on_products() {
		if ( $this->products_exists() ) {
			update_option( self::OPTION_ENABLED_PRODUCTS, true ); // enable Products by default.
			$products_categories = get_categories(
				array(
					'taxonomy'   => 'product_cat',
					'hide_empty' => false,
					'fields'     => 'ids',
				)
			);
			if ( is_array( $products_categories ) ) {
				$this->set_products_categories_checked( $products_categories );
			}
		}
		remove_action( 'init', [ $this, 'maybe_turn_on_products' ] );
	}
	/**
	 * Is scope parameters updated and new content audit required
	 *
	 * @since 0.8.0
	 *
	 * @return bool
	 */
	public function is_scope_updated() {
		return (bool) get_option( self::OPTION_SCOPE_UPDATED, false );
	}
	/**
	 * Reset scope updated parameter.
	 *
	 * @since 0.8.0
	 *
	 * @return void
	 */
	public function reset_scope_updated() {
		update_option( self::OPTION_SCOPE_UPDATED, false );
	}
	/**
	 * Do we have new CPT, that was not approved?
	 *
	 * @since 0.8.0
	 *
	 * @return string[] CPT list without already approved or active items, like 'product'.
	 */
	public function has_new_cpt_for_tip() {
		return array_diff( array_keys( $this->get_custom_post_types( false ) ), (array) get_option( self::OPTION_SCOPE_EXISTING_POST_TYPES, [] ), $this->get_custom_post_types_enabled( true ) );
	}
	/**
	 * Add all existing CPT as approved
	 *
	 * @since 0.8.0
	 *
	 * @return void
	 */
	public function approve_existing_cpt() {
		update_option( self::OPTION_SCOPE_EXISTING_POST_TYPES, array_keys( $this->get_custom_post_types( false ) ) );
	}
}
