<?php
declare(strict_types=1);

namespace ahrefs\AhrefsSeo;

/**
 * Base class for content, implement options get and save.
 */
class Ahrefs_Seo_Content {

	private const OPTION_DISABLED_PAGES     = 'ahrefs-seo-content-disabled-pages';
	private const OPTION_DISABLED_POSTS     = 'ahrefs-seo-content-disabled-posts';
	protected const OPTION_ENABLED_PRODUCTS = 'ahrefs-seo-content-enabled-products';
	private const OPTION_ENABLED_ALL_CPT    = 'ahrefs-seo-content-enabled-all-cpt';
	private const OPTION_POSTS_CAT          = 'ahrefs-seo-content-posts-cat';
	private const OPTION_PAGES_CAT          = 'ahrefs-seo-content-pages-cat';
	private const OPTION_PRODUCTS_CAT       = 'ahrefs-seo-content-products-cat';
	private const OPTION_CPT_LIST           = 'ahrefs-seo-content-cpt-list';
	protected const OPTION_WAITING_WEEKS    = 'ahrefs-seo-content-waiting-weeks';
	protected const DEFAULT_WAITING_WEEKS   = 12;

	/**
	 * Get waiting time, before we analyze a post.
	 *
	 * @return int Number of weeks.
	 */
	public function get_waiting_weeks() : int {
		$result = absint( get_option( self::OPTION_WAITING_WEEKS, self::DEFAULT_WAITING_WEEKS ) );
		if ( $result < 1 ) {
			$result = 1;
		}
		return $result;
	}

	/**
	 * Get checked pages using option value
	 *
	 * @return string[]|null
	 */
	public function get_pages_checked() : ?array {
		$result = get_option( self::OPTION_PAGES_CAT, null );
		return is_array( $result ) ? $result : null;
	}

	/**
	 * Get checked post categories from option
	 *
	 * @return string[]|null
	 */
	public function get_posts_categories_checked() : ?array {
		$result = get_option( self::OPTION_POSTS_CAT, null );
		return is_array( $result ) ? $result : null;
	}

	/**
	 * Get checked products categories from option
	 *
	 * @since 0.8.0
	 *
	 * @return string[]|null
	 */
	public function get_products_categories_checked() : ?array {
		$result = get_option( self::OPTION_PRODUCTS_CAT, null );
		return is_array( $result ) ? $result : null;
	}

	/**
	 * Get enabled by user custom post types
	 *
	 * @since 0.8.0
	 *
	 * @param bool $with_standard_posts Include post, page and product.
	 * @return string[]
	 */
	public function get_custom_post_types_enabled( bool $with_standard_posts = false ) : array {
		$result = get_option( self::OPTION_CPT_LIST, null );
		if ( ! is_array( $result ) ) {
			$result = [];
		}
		if ( $with_standard_posts ) {
			$result[] = 'post';
			$result[] = 'page';
			if ( ! $this->is_disabled_for_products() ) {
				$result[] = 'product';
			}
		}
		return $result;
	}

	public function is_disabled_for_pages() : bool {
		$value = get_option( self::OPTION_DISABLED_PAGES );
		return ! empty( $value );
	}

	public function is_disabled_for_posts() : bool {
		$value = get_option( self::OPTION_DISABLED_POSTS );
		return ! empty( $value );
	}

	/**
	 * Is disabled for product post type
	 *
	 * @since 0.8.0
	 *
	 * @return bool
	 */
	public function is_disabled_for_products() : bool {
		$value = get_option( self::OPTION_ENABLED_PRODUCTS );
		return empty( $value );
	}

	/**
	 * Is disabled for all CPT
	 *
	 * @since 0.8.0
	 *
	 * @return bool
	 */
	public function is_disabled_for_all_cpt() : bool {
		$value = get_option( self::OPTION_ENABLED_ALL_CPT );
		return empty( $value );
	}

	protected function set_waiting_weeks( int $waiting_weeks ) : void {
		update_option( self::OPTION_WAITING_WEEKS, max( 1, $waiting_weeks ) );
	}

	protected function set_disabled_for( bool $pages_disabled, bool $posts_disabled, bool $products_disabled, bool $all_cpt_disabled ) : void {
		update_option( self::OPTION_DISABLED_PAGES, $pages_disabled );
		update_option( self::OPTION_DISABLED_POSTS, $posts_disabled );
		update_option( self::OPTION_ENABLED_PRODUCTS, ! $products_disabled );
		update_option( self::OPTION_ENABLED_ALL_CPT, ! $all_cpt_disabled );
	}

	/**
	 * Set checked pages list
	 *
	 * @param string[] $values
	 * @return void
	 */
	protected function set_pages_checked( ?array $values ) : void {
		if ( ! is_null( $values ) ) {
			update_option( self::OPTION_PAGES_CAT, $values );
		} else {
			delete_option( self::OPTION_PAGES_CAT );
		}
	}

	/**
	 * @param string[] $values
	 * @return void
	 */
	protected function set_posts_categories_checked( array $values ) : void {
		update_option( self::OPTION_POSTS_CAT, $values );
	}

	/**
	 * Set products categories checked
	 *
	 * @since 0.8.0
	 *
	 * @param string[] $values
	 * @return void
	 */
	protected function set_products_categories_checked( array $values ) : void {
		update_option( self::OPTION_PRODUCTS_CAT, $values );
	}

	/**
	 * Set checked custom post types
	 *
	 * @since 0.8.0
	 *
	 * @param string[] $values
	 * @return void
	 */
	protected function set_cpt_checked( array $values ) : void {
		update_option( self::OPTION_CPT_LIST, $values );
	}

	/**
	 * Add pages to checked pages list.
	 * Use pages id, not a post tax.
	 *
	 * @param string[] $post_ids
	 * @return void
	 */
	protected function pages_add_to_checked( array $post_ids ) : void {
		$enabled = $this->get_pages_checked() ?? [];
		$enabled = array_unique( array_merge( $enabled, $post_ids ) );
		$this->set_pages_checked( $enabled );
		if ( count( $enabled ) && $this->is_disabled_for_pages() ) { // enable.
			$this->set_disabled_for( false, $this->is_disabled_for_posts(), $this->is_disabled_for_products(), $this->is_disabled_for_all_cpt() );
		}
	}

	/**
	 * Remove pages from checked pages list
	 *
	 * @param Post_Tax[] $post_taxes
	 * @return void
	 */
	protected function pages_remove_from_checked( array $post_taxes ) : void {
		$enabled  = $this->get_pages_checked() ?? [];
		$post_ids = array_map(
			function( Post_Tax $item ) {
				return $item->get_post_id();
			},
			$post_taxes
		);
		$enabled  = array_unique( array_diff( $enabled, $post_ids ) );
		$this->set_pages_checked( $enabled );
		if ( ! count( $enabled ) && ! $this->is_disabled_for_pages() ) { // disable.
			$this->set_disabled_for( true, $this->is_disabled_for_posts(), $this->is_disabled_for_products(), $this->is_disabled_for_all_cpt() );
		}
	}

	/**
	 * Clear internal cached data
	 */
	public function clear_cache() : void {
		delete_option( self::OPTION_DISABLED_PAGES );
		delete_option( self::OPTION_DISABLED_POSTS );
		delete_option( self::OPTION_PAGES_CAT );
		delete_option( self::OPTION_POSTS_CAT );
	}

	/**
	 * Is product custom post type exists
	 *
	 * @since 0.8.0
	 *
	 * @return bool
	 */
	public function products_exists() : bool {
		return in_array(
			'product',
			(array) get_post_types(
				[
					'public'   => true,
					'_builtin' => false,
				]
			),
			true
		);
	}

}
