<?php
declare(strict_types=1);

namespace ahrefs\AhrefsSeo;

/*
We can have 3 types of keywords by source (db ahrefs_seo_content table, keyword_source field):
- up to 10 proposed by gsc: "gsc" badge in popup dialog;
- up to 5 proposed by tf-idf: "tf-idf" badge;
- assigned manually: input field in popup dialog.

We run bulk keywords update:
- on initial wizard, at last step;
- on each audit (when new snapshot created);

We update keywords of each single post when keyword is not approved by user.

Time:
- When keywords received from GSC we do 2-3 requests (case A below) or 0-1 request (case B below);
- We are doing at pauses between requests (Ahrefs_Seo_Analytics::API_MIN_DELAY, counting from previous request started time);
- GSC can have a big delay before API returned data on first request (it seems it cacne results and next requests are much faster).

How is GSC working now?
A. Common update, when we load detail during content audit or receive updated keywords suggestions (on keywords popup opened):
When we load keyword suggestions from GSC we make up to 3 queries:
- get total clicks, positions and impressions - this required, as we must show values in percents;
- get top 10 results - what we show at the keywords popup. Search current keyword here, if not found then:
- get clicks, positions and impressions exactly for current keyword.

B. Fast update, when user sets a new keyword:
- we try to load position from cached gsc result (from last query)
- if not found, then make single request using filter with current keyword and update position value (do not cache result).
This is working much faster, that common update (with 2 or 3 queries).
But we do not have impressions. clicks and percents values. If "Change" link is clicked again, we will load all the details and update keywords table with fresh results (using case A).


Summary:
We do NOT update keywords using automated results when they already set by user ("is_approved_keyword" is set).
GSC or TF-IDF can propose another keywords in different times (when GSC will have new click details or more new posts added for TF-IDF).

Note: we must add any inactive post (not added at the Wizard or created later) post to content table before assign keywords or use cached suggestions.
Note: the "post" word means any post, page, product, custom post type, category or product category page.
*/

/**
 * Wrapper class for keywords features.
 */
class Ahrefs_Seo_Keywords {
	/**
	 * @var array
	 */
	private $all_posts = [];

	/**
	 * @var array
	 */
	private $all_keywords = [];

	/**
	 * @var Ahrefs_Seo_Keywords
	 */
	private static $instance;

	/**
	 * @var array<int, string>
	 */
	private $keywords_source_cache = [];

	/** Allow create new keywords once per seconds. */
	const KEY_MIN_DELAY = 0.05;
	/** @var float Time when last keyword created. */
	private $last_query_time = 0;

	/**
	 * @var Ahrefs_Seo_Data_Content
	 */
	private $data_content;

	/**
	 * @var Ahrefs_Seo_Analytics
	 */
	private $analytics;

	/**
	 * Return the instance
	 *
	 * @param Ahrefs_Seo_Analytics|null $analytics Analytics instance for use in requests to GSC.
	 * @return Ahrefs_Seo_Keywords
	 */
	public static function get( ?Ahrefs_Seo_Analytics $analytics = null ) : Ahrefs_Seo_Keywords {
		if ( empty( self::$instance ) ) {
			self::$instance = new self( Ahrefs_Seo_Data_Content::get() );
		}
		if ( ! is_null( $analytics ) ) {
			self::$instance->analytics = $analytics;
		}
		if ( is_null( self::$instance->analytics ) ) { // @phpstan-ignore-line
			self::$instance->analytics = Ahrefs_Seo_Analytics::get();
		}
		return self::$instance;
	}

	public function __construct( Ahrefs_Seo_Data_Content $data_content ) {
		$this->data_content = $data_content;
	}

	/**
	 * Get keywords assigned to post
	 *
	 * @param Post_Tax $post_tax
	 * @return string|null Assigned keyword if any.
	 */
	public function post_keyword_get( Post_Tax $post_tax ) : ?string {
		global $wpdb;
		return $wpdb->get_var( $wpdb->prepare( "SELECT keyword FROM {$wpdb->ahrefs_content} WHERE post_id = %d AND snapshot_id = %d AND taxonomy = %s", $post_tax->get_post_id(), $post_tax->get_snapshot_id(), $post_tax->get_taxonomy() ) );
	}

	/**
	 * Is this post keywords is approved?
	 *
	 * @param Post_Tax $post_tax Post ID.
	 * @return bool
	 */
	public function post_keywords_is_approved( Post_Tax $post_tax ) : bool {
		global $wpdb;
		$result = $wpdb->get_var( $wpdb->prepare( "SELECT is_approved_keyword FROM {$wpdb->ahrefs_content} WHERE post_id = %d AND snapshot_id = %d AND taxonomy = %s LIMIT 1", $post_tax->get_post_id(), $post_tax->get_snapshot_id(), $post_tax->get_taxonomy() ) );
		return ! empty( $result );
	}

	/**
	 * Reset keywords need update flag.
	 *
	 * @param Post_Tax $post_tax
	 * @return void
	 */
	public function post_keywords_set_updated( Post_Tax $post_tax ) : void {
		global $wpdb;
		Ahrefs_Seo::breadcrumbs( sprintf( '%s (%s)', __METHOD__, (string) $post_tax ) );
		$sql = $wpdb->prepare( "UPDATE {$wpdb->ahrefs_content} SET keywords_need_update = 0 WHERE post_id = %d AND snapshot_id = %d AND taxonomy = %s", $post_tax->get_post_id(), $post_tax->get_snapshot_id(), $post_tax->get_taxonomy() );
		if ( 0 === $wpdb->query( $sql ) ) { // phpcs:ignore WordPress.DB.PreparedSQL.NotPrepared
			$this->maybe_add_post_and_retry_query( $post_tax, $sql );
		}
	}


	/**
	 * Set post keywords (selected and user keyword).
	 * Update position (using cached results from GSC) or set a flag for an update.
	 * Set flag for keywords update if keyword is null.
	 * Will update article recommended action.
	 *
	 * @param Post_Tax    $post_tax Post ID.
	 * @param string|null $keyword Post keyword.
	 * @param string|null $keyword_manual Keywords from manual input field.
	 * @param bool        $reanalyze_everything Update action for current post and other active posts with same keywords (before and after update).
	 * @return bool
	 */
	public function post_keywords_set( Post_Tax $post_tax, ?string $keyword, ?string $keyword_manual = null, bool $reanalyze_everything = true ) : bool {
		global $wpdb;
		Ahrefs_Seo::breadcrumbs( sprintf( '%s (%s) (%s) (%s) (%d)', __METHOD__, (string) $post_tax, $keyword, $keyword_manual, $reanalyze_everything ) );
		if ( ! is_null( $keyword ) && strlen( $keyword ) > 191 ) {
			$keyword = function_exists( 'mb_substr' ) ? mb_substr( $keyword, 0, 191 ) : substr( $keyword, 0, 191 );
		}

		if ( $post_tax->exists() && $post_tax->is_tax_or_published() ) { // only for existing and published posts.
			$values                    = [];
			$place_holders             = [];
			$position                  = null; // position value for newly set keyword.
			$keyword_old               = $this->post_keyword_get( $post_tax );
			$update_same_keyword_posts = $reanalyze_everything && $keyword_old !== $keyword;

			// need to reanalyze posts with same keyword as current (using keyword before update).
			$posts_reanalyze_old = $update_same_keyword_posts ? Ahrefs_Seo_Advisor::get()->find_active_pages_with_same_keyword( $post_tax ) : [];

			if ( ! is_null( $keyword ) ) {
				// try to update position with fresh value, if exists in cached data.
				$position = $this->load_position_from_cache( $post_tax, $keyword );
			}
			$fields = [
				'keyword'              => $keyword,
				'position'             => $position,
				'position_need_update' => is_null( $position ) ? 1 : 0,
				'keywords_need_update' => is_null( $keyword ) ? 1 : 0,
			];
			$format = [ '%s', '%f', '%d', '%d' ];
			if ( ! is_null( $keyword_manual ) ) {
				$fields['keyword_manual'] = $keyword_manual;
				$format[]                 = '%s';
			}
			$query_result = $wpdb->update(
				$wpdb->ahrefs_content,
				$fields,
				[
					'post_id'     => $post_tax->get_post_id(),
					'taxonomy'    => $post_tax->get_taxonomy(),
					'snapshot_id' => $post_tax->get_snapshot_id(),
				],
				$format,
				[ '%d', '%d' ]
			);
			if ( 0 === $query_result ) {
				$this->maybe_add_post_and_retry_query( $post_tax, $wpdb->last_query );
			} elseif ( false === $query_result ) {
				$this->post_keywords_set_updated( $post_tax );
				Ahrefs_Seo::notify( new Ahrefs_Seo_Exception( sprintf( 'Query failed (keywords set) (%s) (%s) (%s) (%s)', wp_json_encode( func_get_args() ), wp_json_encode( $fields ), $wpdb->last_query, $wpdb->last_error ) ) );
			}

			$content_audit = new Content_Audit_Current( $post_tax->get_snapshot_id() );
			if ( is_null( $position ) && $reanalyze_everything ) {
				// update position immediately using fast update only when reanalyze_everything required. Because this additional API call may be a reason of possible rate error.
				$content_audit->update_post_info_position( [ $post_tax ], true );
			}
			if ( $reanalyze_everything ) {
				// update recommended action.
				$content_audit->reanalyze_post( $post_tax );
			}

			// need to reanalyze posts with same keyword as current (using keyword after update).
			$posts_reanalyze_new = $update_same_keyword_posts && '' !== $keyword ? Ahrefs_Seo_Advisor::get()->find_active_pages_with_same_keyword( $post_tax ) : [];

			$posts_reanalyze = array_merge( $posts_reanalyze_old ?? [], $posts_reanalyze_new ?? [] );

			if ( ! empty( $posts_reanalyze ) ) {
				$posts_reanalyze = array_unique( $posts_reanalyze );
				foreach ( $posts_reanalyze as $post_tax ) {
					// update recommended action.
					$content_audit->reanalyze_post( $post_tax );
				}
			}

			// if this is current snapshot and new snapshot exists: update it with cached suggestions (do not do request to API) and current keyword.
			$new_snapshot_id = ( new Content_Audit() )->get_snapshot_id();
			if ( ! is_null( $new_snapshot_id ) && ( new Snapshot() )->has_current_and_new_snapshots() && $post_tax->get_snapshot_id() !== $new_snapshot_id ) {
				$post_tax_kw = Post_Tax_With_Keywords::create_from( $post_tax )->load_from_db(); // load existing cached suggestions.
				$post_tax_kw->set_snapshot_id( $new_snapshot_id ); // assign new snapshot.
				$this->set_cached_suggestions( $post_tax_kw ); // and save to new snapshot.
				// also update everything in 'new' snapshot.
				$this->post_keywords_set( $post_tax->set_snapshot_id( $new_snapshot_id ), $keyword, $keyword_manual, $reanalyze_everything );
			}

			return true;
		} else {
			// todo: maybe delete post from content audit table.
		}
		return false;
	}

	/**
	 * Check, add new inactive post to content table and retry initial sql query.
	 *
	 * Newly added posts or initially inactive posts are missing at the content table.
	 * Must add post before use it, otherwise update keywords or cached suggestions do not work.
	 *
	 * @param Post_Tax $post_tax
	 * @param string   $sql
	 *
	 * @return void
	 */
	private function maybe_add_post_and_retry_query( Post_Tax $post_tax, string $sql ) : void {
		global $wpdb;
		Ahrefs_Seo::breadcrumbs( __METHOD__ . ' ' . (string) $post_tax . " ($sql)" );
		$error = $wpdb->last_error;
		if ( $error ) { // is there any error?
			// maybe this post was not added to content table at all?
			if ( is_null( $wpdb->get_var( $wpdb->prepare( "SELECT post_id FROM {$wpdb->ahrefs_content} WHERE post_id = %d AND snapshot_id = %d AND taxonomy = %s LIMIT 1", $post_tax->get_post_id(), $post_tax->get_snapshot_id(), $post_tax->get_taxonomy() ) ) ) ) {
				Ahrefs_Seo_Data_Content::get()->add_post_as_added_since_last( $post_tax->get_post_id(), $post_tax->get_taxonomy() );

				// retry query.
				if ( '' !== $sql && 0 === $wpdb->query( $sql ) ) { // phpcs:ignore WordPress.DB.PreparedSQL.NotPrepared -- we already tried to execute this sql code before, this time we just retry it.
					Ahrefs_Seo::notify( new \Exception( sprintf( 'SQL query failed again (%s) (%s) (%s)', $sql, $error, $wpdb->last_error ) ) );
				}
			} else {
				Ahrefs_Seo::notify( new \Exception( sprintf( 'SQL query failed (%s) (%s)', $sql, $error ) ) );
			}
		}
	}

	/**
	 * Create keywords for single post/term.
	 * Wrapper with delay around Keywords_Search calls.
	 *
	 * @param Post_Tax $post_tax
	 * @param int      $keywords_limit
	 * @return Data_Keyword_Content_Storage|null Keywords for the post Data_Keyword_Content_Storage or null.
	 */
	private function create_single_post_keywords( Post_Tax $post_tax, int $keywords_limit = 10 ) : ?Data_Keyword_Content_Storage {
		if ( function_exists( 'wp_raise_memory_limit' ) ) {
			wp_raise_memory_limit( 'admin' );
		}
		unset( $this->all_keywords );
		$this->maybe_do_a_pause();
		// run research.
		return ( new Keywords_Search( $post_tax, $keywords_limit ) )->get_all_keywords();
	}

	/**
	 * Do a minimal delay between requests.
	 * Used to prevent 504 errors.
	 */
	private function maybe_do_a_pause() : void {
		$time_since = microtime( true ) - $this->last_query_time;
		if ( $time_since < self::KEY_MIN_DELAY && ! defined( 'AHREFS_SEO_IGNORE_DELAY' ) ) {
			Ahrefs_Seo::usleep( intval( ceil( self::KEY_MIN_DELAY - $time_since ) * 1000000 ) );
		}
		$this->last_query_time = microtime( true );
	}

	/**
	 * Get cached details for post
	 *
	 * @param Post_Tax $post_tax
	 *
	 * @return Post_Tax_With_Keywords
	 */
	protected function get_cached_detail_for_post( Post_Tax $post_tax ) : Post_Tax_With_Keywords {
		$result = Post_Tax_With_Keywords::create_from( $post_tax );
		if ( $post_tax->is_tax_or_published() ) { // only for existing and published posts or taxonomy items.
			$result->load_from_db();
		}
		return $result;
	}

	/**
	 * Set cached suggestions
	 *
	 * @param Post_Tax_With_Keywords $post_tax_kw
	 * @return void
	 */
	private function set_cached_suggestions( Post_Tax_With_Keywords $post_tax_kw ) : void {
		global $wpdb;
		$keywords          = $post_tax_kw->get_keywords();
		$keywords2         = $post_tax_kw->get_keywords2();
		$keywords2_low_len = $post_tax_kw->get_keywords2_low_len();
		$keywords_pos      = $post_tax_kw->get_keywords_pos();
		if ( isset( $keywords['error'] ) ) {
			unset( $keywords['error'] ); // it make no sense to cache errors.
		}
		$sql = $wpdb->prepare(
			"UPDATE {$wpdb->ahrefs_content} SET kw_gsc = %s, kw_idf = %s, kw_pos = %s, kw_low = %d, updated=updated WHERE post_id = %d AND snapshot_id = %d AND taxonomy = %s", // do not change 'updated' value.
			wp_json_encode( $keywords ),
			wp_json_encode( $keywords2 ),
			wp_json_encode( $keywords_pos ),
			$keywords2_low_len,
			$post_tax_kw->get_post_id(),
			$post_tax_kw->get_snapshot_id(),
			$post_tax_kw->get_taxonomy()
		);
		if ( 0 === $wpdb->query( $sql ) ) { // phpcs:ignore WordPress.DB.PreparedSQL.NotPrepared
			$this->maybe_add_post_and_retry_query( $post_tax_kw, $sql );
		}
	}

	/**
	 * Get current keyword and keywords data from GSC or TF-IDF.
	 *
	 * @param Post_Tax[] $post_taxes Post ID list.
	 * @param int        $limit
	 * @param bool       $without_totals Do not make additional query for total values.
	 * @param bool       $gsc_only Do not use tf-idf.
	 *
	 * @return Post_Tax_With_Keywords[]
	 */
	public function get_full_detail_for_posts( array $post_taxes, ?int $limit = null, bool $without_totals = false, $gsc_only = false ) : array {
		$v    = func_get_args();
		$v[0] = Post_Tax::id( $post_taxes );
		Ahrefs_Seo::breadcrumbs( sprintf( '%s %s', __METHOD__, wp_json_encode( $v ) ) );
		$start_date = '';
		$end_date   = '';
		$this->get_time_period_keywords( $start_date, $end_date );
		$days_count = 0;

		$results          = [];
		$requests         = [];
		$request_keywords = [];

		foreach ( $post_taxes as $post_tax ) {
			$post_tax_kw = Post_Tax_With_Keywords::create_from( $post_tax )->load_from_db();
			if ( $post_tax->exists() && $post_tax->is_tax_or_published() && ( '' !== $post_tax->get_url() ) ) { // only for existing and published posts or terms with non empty url.
				$post_tax_string = (string) $post_tax_kw;
				$url             = apply_filters( 'ahrefs_seo_search_traffic_url', $post_tax_kw->get_url() );

				if ( $this->analytics->is_gsc_set() ) {
					$request_keywords[ $post_tax_string ] = $post_tax_kw->get_keyword_curent();
					$requests[ (string) $post_tax_kw ]    = $url; // will load suggestions from GSC.
				} else {
					$post_tax_kw->set_error_message( 'Google Search Console is not connected.' );
				}
				$post_tax_kw->set_keywords2( $this->create_single_post_keywords( $post_tax_kw, 10 ) ); // do not apply $limit parameter here.
				$this->set_cached_suggestions( $post_tax_kw );
			} else {
				$post_tax_kw->set_error_message( 'This page cannot be found. It is possible that you’ve archived the page or changed the page ID. Please reload the page & try again.' );
			}
			$results[ (string) $post_tax_kw ] = $post_tax_kw; // results without fresh GSC data.
		}
		// make single call for all pending urls.
		if ( ! empty( $requests ) ) {
			$clicks = $this->analytics->get_clicks_and_impressions_by_urls( $requests, $start_date, $end_date, $limit, $without_totals, $request_keywords );
			// update results for each post id.
			foreach ( $requests as $post_tax_string => $url ) {
				$clicks_of_url = is_array( $clicks ) ? ( $clicks[ $post_tax_string ] ?? null ) : null;
				if ( ! empty( $clicks_of_url ) ) {
					if ( ! empty( $clicks_of_url['kw_pos'] ) ) {
						$results[ $post_tax_string ]->set_keywords_pos( $clicks_of_url['kw_pos'] );
					}
					$results[ $post_tax_string ]->set_keywords( $clicks_of_url );
					if ( ! empty( $clicks_of_url['error'] ) ) {
						if ( $clicks_of_url['error'] instanceof \Exception ) {
							$clicks_of_url['error'] = $clicks_of_url['error']->getMessage();
						}
						if ( is_string( $clicks_of_url['error'] ) ) {
							$results[ $post_tax_string ]->set_error_message( $clicks_of_url['error'] );
						}
					}
				} elseif ( ! $gsc_only ) {
					$error = $this->analytics->get_message();
					if ( $error ) {
						$results[ $post_tax_string ]->set_error_message( $error );
					}
				}
				// set fresh cached values.
				$this->set_cached_suggestions( $results[ $post_tax_string ] );
			}
		}
		return $results;
	}

	/**
	 * Fill keyword row using cached GSC results
	 *
	 * @param array $cached_results
	 * @param array $keyword_row
	 * @return void
	 */
	private function fill_using_cached_gsc_results( array $cached_results, array &$keyword_row ) : void {
		$keyword = $keyword_row[1];
		if ( ! empty( $cached_results ) && ! is_null( $keyword ) ) {
			foreach ( $cached_results as $item ) {
				if ( isset( $item['query'] ) && 0 === strcasecmp( $keyword, $item['query'] ) ) {
					$keyword_row[3] = $item['pos'] ?? null;
					$keyword_row[4] = $item['clicks'] ?? null;
					$keyword_row[5] = $item['impr'] ?? null;
				}
			}
		}
	}

	/**
	 * Return the same data, as 'keywords-list' template expected.
	 *
	 * @param Post_Tax $post_tax
	 * @param bool     $use_cached_data
	 * @return array<string, mixed>
	 */
	public function get_suggestions( Post_Tax $post_tax, bool $use_cached_data = false ) : array {
		$errors   = null;
		$keywords = [];
		if ( ! $use_cached_data ) {
			$data        = $this->get_full_detail_for_posts( [ $post_tax ] );
			$post_tax_kw = $data[ (string) $post_tax ];
		} else {
			$post_tax_kw = $this->get_cached_detail_for_post( $post_tax );
			if ( empty( $post_tax_kw->get_keywords2() ) ) { // always fill TF-IDF keywords, even this is cached results.
				$post_tax_kw->set_keywords2( $this->create_single_post_keywords( $post_tax_kw, 10 ) );
			}
		}
		$source_gsc            = false; // source of report is GSC data.
		$total_clicks          = 0;
		$total_impr            = 0;
		$keyword_manual        = $post_tax_kw->get_keyword_manual();
		$cached_positions      = [];
		$no_suggestions_found  = false;
		$no_suggestions_cached = false;
		$data_keywords         = $post_tax_kw->get_keywords();
		$data_keywords2        = $post_tax_kw->get_keywords2();
		$keyword               = $post_tax_kw->get_keyword_curent() ?? '';
		if ( ! empty( $data_keywords ) && ! empty( $data_keywords['result'] ) ) {
			$cached_positions        = $data_keywords['result'];
			$data_keywords['result'] = array_slice( $data_keywords['result'], 0, 10 ); // use only top 10 items, ordered by clicks.
			$source_gsc              = true;
			$total_clicks            = ! empty( $data_keywords['total_clicks'] ) ? $data_keywords['total_clicks'] : 1;
			$total_impr              = ! empty( $data_keywords['total_impr'] ) ? $data_keywords['total_impr'] : 1;
			foreach ( $data_keywords['result'] as $item ) {
				if ( isset( $item['query'] ) && is_string( $item['query'] ) ) {
					$keywords[] = [ 0 === strcasecmp( $item['query'], $keyword ), $item['query'], 'gsc', $item['pos'], $item['clicks'], $item['impr'] ];
				}
			}
		}
		// is some keyword already selected? We want to select only single item.
		$item_selected = array_sum(
			array_map(
				function( $item ) {
					return $item[0] ? 1 : 0;
				},
				$keywords
			)
		) > 0;
		if ( ! empty( $data_keywords2 ) ) {
			if ( 0 === $total_clicks ) {
				$total_clicks = 1;
			}
			if ( 0 === $total_impr ) {
				$total_impr = 1;
			}
			$data_keywords2 = array_slice( $data_keywords2, 0, 5 ); // use only top 5 items.
			foreach ( $data_keywords2 as $item ) {
				$word = $item->get_keyword();
				$row  = [ ! $item_selected && 0 === strcasecmp( $word, $keyword ), $word, 'tf-idf', null, '-', '-' ];
				$this->fill_using_cached_gsc_results( $cached_positions, $row );
				$keywords[] = $row;
			}
		}
		$item_selected = array_sum(
			array_map(
				function( $item ) {
					return $item[0] ? 1 : 0;
				},
				$keywords
			)
		) > 0;

		// do we have currently selected item somehere? Keyword is set, but we did not find it at GSC or TF-IDF suggestions and it is not a manual keyword too...
		if ( ( '' !== $keyword ) && ! $item_selected && ( is_null( $keyword_manual ) || 0 !== strcasecmp( $keyword_manual, $keyword ) ) ) {
			// replace manual keyword by selected keyword.
			$keyword_manual = $keyword;
		}

		$item = [ ! $item_selected && ! is_null( $keyword_manual ) && 0 === strcasecmp( $keyword_manual, $keyword ), $keyword_manual ?? '', 'manual', null, '-', '-' ];
		$this->fill_using_cached_gsc_results( $cached_positions, $item );

		$keywords[] = $item;
		if ( empty( $keywords ) ) {
			// show errors.
			if ( ! empty( $post_tax_kw->get_error_message() ) ) {
				$errors = $post_tax_kw->get_error_message();
			} else {
				// no keyword suggestions found.
				$no_suggestions_found = true;
			}
		}

		if ( ( is_null( $post_tax_kw->get_keyword_curent() ) || is_null( $post_tax_kw->get_keywords() ) ) && is_null( $post_tax_kw->get_keywords2() ) ) {
			$no_suggestions_cached = true;
		}

		if ( ! $use_cached_data && ! empty( $post_tax_kw->get_error_message() ) ) {
			$errors = $post_tax_kw->get_error_message();
		}

		return [
			'post_id'      => (string) $post_tax_kw, // post_tax_string.
			'keyword'      => $post_tax_kw->get_keyword_curent() ?? [], // value returned from from cache can be null.
			'keywords'     => $keywords,
			'total_clicks' => $total_clicks,
			'total_impr'   => $total_impr,
			'errors'       => $errors,
			'post_tax'     => $post_tax_kw,
		];
	}

	/**
	 * Get time period
	 *
	 * @param string $start_date
	 * @param string $end_date
	 * @return void
	 */
	public function get_time_period_keywords( string &$start_date, string &$end_date ) : void {
		$start_date = date( 'Y-m-d', strtotime( sprintf( '- 3 month' ) ) );
		$end_date   = date( 'Y-m-d' );
	}

	/**
	 * Load position of keyword.
	 *
	 * @param Post_Tax[] $post_taxes Items to load current keyword position for.
	 * @return Post_Tax_With_Keywords[]
	 */
	public function load_position_value_fast( array $post_taxes ) : array {
		$post_taxes_kw = [];
		foreach ( $post_taxes as $post_tax ) {
			$post_taxes_kw[] = Post_Tax_With_Keywords::create_from( $post_tax )->load_from_db();
		}
		$url_with_keywords = [];
		/** @var Post_Tax_With_Keywords[] $url_to_item */
		$url_to_item = [];
		array_walk(
			$post_taxes_kw,
			function( Post_Tax_With_Keywords &$post_tax_kw ) use ( &$url_with_keywords, &$url_to_item ) {
				$url             = $post_tax_kw->get_url();
				$keyword_current = $post_tax_kw->get_keyword_curent();
				if ( ! empty( $keyword_current ) && ! empty( $url ) ) {
					$url_with_keywords[ $url ] = $keyword_current;
					$url_to_item[ $url ]       = &$post_tax_kw;
					$post_tax_kw->set_position( -1 ); // set error, will overwrite it with value from API later.
				} else {
					$post_tax_kw->set_position( Ahrefs_Seo_Data_Content::POSITION_MAX );
				}
			}
		);
		if ( $url_with_keywords ) { // have something to load?
			$rows = $this->analytics->get_position_fast( $url_with_keywords );
			if ( $rows ) {
				foreach ( $rows as $url => $position ) { // fill with received values.
					$url_to_item[ $url ]->set_position( floatval( $position['pos'] ) )->set_keywords_pos( $position );
				}
			}
		}

		return $post_taxes_kw;
	}

	/**
	 * Return position for URL and current keyword.
	 * Load top 10 results for the page and result for current keyword.
	 *
	 * @param Post_Tax[] $post_taxes
	 * @return Post_Tax_With_Keywords[]
	 */
	public function load_position_value( array $post_taxes ) : array {
		$post_taxes_with_kw = $this->get_full_detail_for_posts( $post_taxes, 10, true, true );
		foreach ( $post_taxes_with_kw as &$post_tax_kw ) {
			$current_keyword_lower = function_exists( 'mb_strtolower' ) ? mb_strtolower( $post_tax_kw->get_keyword_curent() ?? '' ) : strtolower( $post_tax_kw->get_keyword_curent() ?? '' );
			$keywords              = $post_tax_kw->get_keywords();
			if ( ! empty( $keywords ) && ! empty( $keywords['result'] ) ) {
				foreach ( $keywords['result'] as $item ) {
					if ( isset( $item['query'] ) && is_string( $item['query'] ) && ( function_exists( 'mb_strtolower' ) ? mb_strtolower( $item['query'] ) : strtolower( $item['query'] ) ) === $current_keyword_lower ) {
							$post_tax_kw->set_position( floatval( $item['pos'] ) )->set_keywords_pos( $item );
						break;
					}
				}
			}
		}
		return $post_taxes_with_kw;
	}

	/**
	 * Load position of keyword from GSC cache & GSG current keyword position.
	 *
	 * @param Post_Tax $post_tax
	 * @param string   $current_keyword
	 * @return float|null Position or null.
	 */
	public function load_position_from_cache( Post_Tax $post_tax, string $current_keyword ) : ?float {
		$result = null;
		$cached = Post_Tax_With_Keywords::create_from( $post_tax )->load_from_db();
		$kw_gsc = $cached->get_keywords();
		$kw_pos = $cached->get_keywords_pos();
		$items  = ! is_null( $kw_gsc ) && ! empty( $kw_gsc['result'] ) && is_array( $kw_gsc['result'] ) ? $kw_gsc['result'] : [];
		if ( ! is_null( $kw_pos ) ) {
			$items[] = $kw_pos;
		}

		$keyword_lower = strtolower( $current_keyword );
		foreach ( $items as $item ) {
			if ( isset( $item['query'] ) && strtolower( $item['query'] ) === $keyword_lower ) {
				$result = isset( $item['pos'] ) ? floatval( $item['pos'] ) : null;
				break;
			}
		}

		return $result;
	}

}
