<?php
declare(strict_types=1);
namespace ahrefs\AhrefsSeo;

/**
 * Helper for work with content
 *
 * @since 0.8.2
 */
class Helper_Content {

	public const LANG_PREFIX = 'lng:';

	public static function get() : self {
		return new self();
	}

	/**
	 * Get post permalink
	 *
	 * @param int $post_id
	 * @return string
	 */
	public function get_post_permalink( int $post_id ) : string {
		return get_permalink( $post_id ) ?: '';
	}

	/**
	 * Get category permalink
	 *
	 * @param int $term_id
	 * @return string
	 */
	public function get_category_link( int $term_id ) : string {
		return get_category_link( $term_id );
	}

	/**
	 * Get term permalink
	 *
	 * @param int    $term_id
	 * @param string $taxonomy
	 * @return string
	 */
	public function get_term_link( int $term_id, string $taxonomy = '' ) : string {
		$result = get_term_link( $term_id, $taxonomy );
		return is_string( $result ) ? $result : '';
	}

	/**
	 * Get post categories or product categories list
	 *
	 * @param string $taxonomy
	 * @param array  $args
	 * @param string $prefix
	 * @param int    $limit
	 * @return array<\WP_Term|\stdClass>
	 */
	public function get_categories( string $taxonomy, array $args, string $prefix, ?int $limit = null ) : array {
		$args['taxonomy'] = $taxonomy;
		if ( ! is_null( $limit ) ) {
			$args['number'] = $limit;
		}
		$result = get_terms( $args );
		return is_array( $result ) ? $result : [];
	}

	/**
	 * Get currently selected language.
	 *
	 * @return string|null
	 */
	public function get_lang() : ?string {
		return null;
	}

	/**
	 * Show taxonomy terms checklist
	 *
	 * @param string         $taxonomy
	 * @param int[]|string[] $selected_cats
	 * @return void
	 */
	public function terms_checklists( string $taxonomy, array $selected_cats ) : void {
		wp_terms_checklist(
			0,
			[
				'taxonomy'             => $taxonomy,
				'descendants_and_self' => 0,
				'selected_cats'        => $selected_cats,
				'checked_ontop'        => false,
			]
		);
	}

	/**
	 * Get all existing taxonomy terms
	 *
	 * @param string $taxonomy
	 * @return int[] List of term id.
	 */
	public function get_all_term_ids( string $taxonomy ) : array {
		$args = [
			'taxonomy'   => $taxonomy,
			'fields'     => 'ids',
			'hide_empty' => false,
		];
		/** @var int[]|\WP_Error */
		$result = get_terms( $args );
		return is_array( $result ) ? $result : [];
	}

	/**
	 * Get posts list
	 *
	 * @param array $args
	 * @return \WP_Post[]|int[]
	 */
	public function get_posts( array $args ) : array {
		$result = get_posts( $args );
		return is_wp_error( $result ) ? [] : $result;
	}

	/**
	 * Displays a categories drop-down for filtering on the Posts list table.
	 *
	 * @param string $current_post_type
	 * @param string $current_category
	 * @param string $cat_value_raw
	 * @param bool   $is_lang_used
	 * @return void
	 */
	public function categories_dropdown( string $current_post_type = '', string $current_category = '', string $cat_value_raw = '', bool $is_lang_used = false ) : void {
		$tax = get_taxonomy( 'category' );
		$d   = Ahrefs_Seo_Data_Content::CAT_FILTER_DIVIDER; // short name for divider.

		$dropdown_options = array(
			'show_option_all' => ( $tax instanceof \WP_Taxonomy ) && property_exists( $tax->labels, 'all_items' ) ? $tax->labels->all_items : '',
			'hide_empty'      => 0,
			'hierarchical'    => 1,
			'show_count'      => 0,
			'orderby'         => 'name',
		);
		unset( $tax );
		if ( in_array( $current_post_type, [ 'post', 'page' ], true ) && ! empty( $current_category ) ) {
			$dropdown_options['selected'] = $current_category;
		}
		?>
		<label class="screen-reader-text" for="cat"><?php __( 'Filter by category' ); ?></label>
		<select name="cat" id="cat" class="postform">
			<?php
			$items         = [];
			$item          = new \stdClass();
			$item->term_id = '000';
			$item->name    = 'All Categories';
			$item->parent  = '0';
			$items[]       = $item;

			// Posts.
			$item          = new \stdClass();
			$item->term_id = "cat{$d}0";
			$item->name    = 'Posts';
			$item->parent  = '0';
			$items[]       = $item;
			$categories    = get_terms( 'category', $dropdown_options );
			if ( is_array( $categories ) ) {
				/** \WP_Term[]|\StdClass $categories */
				foreach ( $categories as $cat ) {
					$item          = new \stdClass();
					$item->term_id = "cat{$d}" . $cat->term_id;
					$item->name    = $cat->name;
					$item->count   = $cat->count;
					$item->parent  = "cat{$d}" . $cat->parent;
					$items[]       = $item;
				}
			}
			// Pages.
			$content = new Ahrefs_Seo_Content_Settings();
			$pages   = $content->get_pages_list();

			$item          = new \stdClass();
			$item->term_id = "page{$d}0";
			$item->name    = 'Pages';
			$item->parent  = '0';
			$items[]       = $item;

			// Products.
			if ( $content->products_exists() ) {
				$item          = new \stdClass();
				$item->term_id = "product{$d}0";
				$item->name    = 'Products';
				$item->parent  = '0';

				$items[] = $item;

				unset( $dropdown_options['selected'] );
				if ( 'product' === $current_post_type && ! empty( $current_category ) ) {
					$dropdown_options['selected'] = $current_category;
				}

				$categories = get_terms( 'product_cat', $dropdown_options );
				if ( is_array( $categories ) ) {
					/** \WP_Term[] $categories */
					foreach ( $categories as $cat ) {
						$item          = new \stdClass();
						$item->term_id = "product{$d}" . $cat->term_id;
						$item->name    = $cat->name;
						$item->count   = $cat->count;
						$item->parent  = "product{$d}" . $cat->parent;
						$items[]       = $item;
					}
				}
			}

			// All existing public custom post types.
			$all_cpt = ( new Ahrefs_Seo_Content_Settings() )->get_custom_post_types();
			foreach ( $all_cpt as $post_type => $title ) {
				$item          = new \stdClass();
				$item->term_id = "{$post_type}{$d}0";
				$item->name    = $title;
				$item->parent  = '0';
				$items[]       = $item;
			}

			// All Category Pages (all items from taxonomies).
			$item          = new \stdClass();
			$item->term_id = "tax:{$d}0";
			$item->name    = 'All Category Pages';
			$item->parent  = '0';
			$items[]       = $item;

			if ( $content->products_exists() ) { // show subitems only when it make sense.
				// Category (all items from taxonomy Category).
				$item          = new \stdClass();
				$item->term_id = "tax:category{$d}0";
				$item->name    = 'Post Category Pages';
				$item->parent  = "tax:{$d}0";
				$items[]       = $item;

				// Products (all items from taxonomy Product Category).
				$item          = new \stdClass();
				$item->term_id = "tax:product_cat{$d}0";
				$item->name    = 'Product Category Pages';
				$item->parent  = "tax:{$d}0";
				$items[]       = $item;
			}

			$r = [
				'depth'       => 0,
				'orderby'     => 'id',
				'order'       => 'ASC',
				'show_count'  => 0,
				'selected'    => $cat_value_raw,
				'name'        => 'cat',
				'id'          => '',
				'class'       => 'postform',
				'tab_index'   => 0,
				'value_field' => 'term_id',
			];
			echo walk_category_dropdown_tree( $items, 0, $r ); // phpcs:ignore WordPress.XSS.EscapeOutput.OutputNotEscaped
			?>
		</select>
		<?php
	}
}
