<?php
declare(strict_types=1);

namespace ahrefs\AhrefsSeo;

/**
 * Class for work with Post or Taxonomy record from content audit table.
 *
 * @since 0.8.0
 */
class Post_Tax {

	private const SEPARATOR = '|';

	/**
	 * Post ID or term id.
	 *
	 * @var int
	 */
	protected $post_id = 0;
	/**
	 * Taxonomy name or empty string
	 *
	 * @var string
	 */
	protected $taxonomy = '';
	/**
	 * Snapshot ID, if specified
	 *
	 * @var int|null
	 */
	protected $snapshot_id = null;

	/**
	 * Constructor.
	 *
	 * @param int      $post_id
	 * @param string   $taxonomy
	 * @param int|null $snapshot_id
	 */
	public function __construct( int $post_id, string $taxonomy = '', ?int $snapshot_id = null ) {
		$this->post_id  = $post_id;
		$this->taxonomy = $taxonomy;
		if ( ! is_null( $snapshot_id ) ) {
			$this->snapshot_id = $snapshot_id;
		}
	}

	/**
	 * Return post tax string
	 *
	 * @return string "post_id|taxonomy|snapshot_id", where "|" as a separator (class const).
	 */
	public function __toString() : string {
		return (string) $this->post_id . self::SEPARATOR . $this->taxonomy . self::SEPARATOR . (string) $this->snapshot_id;
	}

	/**
	 * Is it a taxonomy term?
	 *
	 * @param string|null $taxonomy Check for exactly this taxonomy.
	 * @return bool Is a same taxonomy as $taxonomy parameter or any taxonomy if parameter is null.
	 */
	public function is_taxonomy( ?string $taxonomy = null ) : bool {
		if ( is_null( $taxonomy ) ) {
			return '' !== $this->taxonomy;
		}
		return '' !== $this->taxonomy && $taxonomy === $this->taxonomy;
	}

	/**
	 * Is it a some post type item?
	 *
	 * @param string|null $post_type Check for exactly this post type.
	 * @return bool Is a same post type as $post_type parameter or any standard or custom post if parameter is null.
	 */
	public function is_post( ?string $post_type = null ) : bool {
		if ( is_null( $post_type ) ) {
			return '' === $this->taxonomy;
		}
		return '' === $this->taxonomy && $post_type === $this->get_post_type();
	}

	/**
	 * Get post or term ID
	 *
	 * @return int
	 */
	public function get_post_id() : int {
		return $this->post_id;
	}

	/**
	 * Get taxonomy
	 *
	 * @return string Taxonomy or empty string for posts.
	 */
	public function get_taxonomy() : string {
		return $this->taxonomy;
	}

	/**
	 * Get snapshot id
	 *
	 * @return int|null Snapshot id or null if it was not set.
	 */
	public function get_snapshot_id() : ?int {
		return $this->snapshot_id;
	}

	/**
	 * Set snapshot id
	 *
	 * @param int $snapshot_id
	 * @return Post_Tax
	 */
	public function set_snapshot_id( int $snapshot_id ) : self {
		$this->snapshot_id = $snapshot_id;
		return $this;
	}

	/**
	 * Get post type
	 *
	 * @return string Post type (post, page, product, etc) or empty string for taxonomy term.
	 */
	public function get_post_type() : string {
		if ( $this->is_post() ) {
			return get_post_type( $this->post_id ) ?: '';
		}
		return '';
	}

	/**
	 * Get post status, if it is a post
	 *
	 * @return string Post status like 'publish' or empty string for taxonomy term.
	 */
	public function get_post_status() : string {
		if ( $this->is_post() ) {
			return get_post_status( $this->post_id ) ?: '';
		}
		return '';
	}

	/**
	 * Is taxonomy a term or a published post.
	 * Вo not check that term really exists.
	 *
	 * @see exists() method for terms.
	 *
	 * @return bool
	 */
	public function is_tax_or_published() : bool {
		return $this->is_taxonomy() || ( 'publish' === $this->get_post_status() );
	}

	/**
	 * This post or term exists
	 *
	 * @return bool
	 */
	public function exists() : bool {
		if ( ! empty( $this->post_id ) ) {
			if ( $this->is_post() ) {
				$post = get_post( $this->get_post_id() );
				return ( $post instanceof \WP_Post ) && ( $post->ID > 0 );
			} else {
				return get_term_field( 'term_id', $this->post_id, $this->taxonomy ) === $this->post_id;
			}
		}
		return false;
	}

	/**
	 * Get url for view post or category/term.
	 * No need to check is_post_type_viewable() or is_taxonomy_viewable() because we work only with publicity viewable post types
	 * and predefined taxonomies (both category and product_cat also are viewable).
	 *
	 * @return string
	 */
	public function get_url() : string {
		$result = '';
		if ( $this->is_post() ) {
			$result = Helper_Content::get()->get_post_permalink( (int) $this->post_id );
		} elseif ( $this->is_taxonomy( 'category' ) ) {
			$result = Helper_Content::get()->get_category_link( $this->post_id );
		} else {
			$result = Helper_Content::get()->get_term_link( (int) $this->post_id );
		}
		return $result;
	}

	/**
	 * Get url for edit post or category/term.
	 *
	 * @see user_can_edit() Is current user able to edit post/term.
	 *
	 * @return string
	 */
	public function get_url_edit() : string {
		if ( $this->is_post() ) {
			return get_edit_post_link( $this->post_id ) ?? '';
		} else {
			$result = get_edit_term_link( $this->post_id, $this->taxonomy );
			return is_string( $result ) ? $result : '';
		}
	}

	/**
	 * Get title of post or category/term.
	 *
	 * @param bool $replace_empty_title Return "Untitled" for items with empty title.
	 * @return string
	 */
	public function get_title( bool $replace_empty_title = false ) : string {
		global $wpdb;
		$result = $wpdb->get_var( $wpdb->prepare( "SELECT title FROM {$wpdb->ahrefs_content} WHERE post_id = %d AND taxonomy = %s AND snapshot_id = %d", $this->post_id, $this->taxonomy, $this->snapshot_id ) );
		if ( is_null( $result ) ) { // fallback.
			if ( $this->is_post() ) {
				$result = get_the_title( $this->post_id );
			} else {
				$result = get_term_field( 'name', (int) $this->post_id, $this->taxonomy );
				$result = is_string( $result ) ? $result : '';
			}
		}
		if ( $replace_empty_title && '' === $result ) {
			return 'Untitled';
		}
		return $result;
	}

	/**
	 * Return where part for sql query with snapshot_id, post_id, taxonomy.
	 *
	 * @return array<string, mixed> Associative array with keys snapshot_id, post_id, taxonomy and their values.
	 */
	public function as_where_array() : array {
		return [
			'snapshot_id' => $this->snapshot_id,
			'post_id'     => $this->post_id,
			'taxonomy'    => $this->taxonomy,
		];
	}

	/**
	 * Return format for where part of sql query with format placeholders.
	 *
	 * @return string[] Placeholders for snapshot_id, post_id, taxonomy.
	 */
	public function as_where_format() : array {
		return [ '%d', '%d', '%s' ];
	}

	/**
	 * Return content of the post or term description.
	 *
	 * @return string
	 */
	public function load_content() : string {
		if ( $this->is_post() ) {
			/** @var \WP_Post|null */
			$post = get_post( $this->post_id );
			return ! is_null( $post ) ? ( $post->post_content ? (string) $post->post_content : (string) $post->post_excerpt ) : '';
		} else {
			$result = get_term_field( 'description', (int) $this->post_id, $this->taxonomy );
			return is_string( $result ) ? $result : '';
		}
	}

	/**
	 * Get recommended action for item.
	 * Read DB.
	 *
	 * @return string|null
	 */
	public function load_action() : ?string {
		global $wpdb;
		return $wpdb->get_var( $wpdb->prepare( "SELECT action FROM {$wpdb->ahrefs_content} WHERE snapshot_id = %d AND post_id = %d AND taxonomy = %s", $this->snapshot_id, $this->post_id, $this->taxonomy ) );
	}

	/**
	 * Current user can edit this post or category/term
	 *
	 * @return bool
	 */
	public function user_can_edit() : bool {
		return ( '' === $this->taxonomy ) && current_user_can( 'edit_post', $this->post_id ) || ( '' !== $this->taxonomy ) && current_user_can( 'manage_categories' );
	}

	/**
	 * Create instance from post tax string
	 *
	 * @param string $post_tax_string
	 * @return Post_Tax
	 */
	public static function create_from_string( string $post_tax_string ) : self {
		if ( strpos( $post_tax_string, self::SEPARATOR ) ) {
			list( $post_id, $taxonomy1, $snapshot_id ) = explode( self::SEPARATOR, $post_tax_string, 3 );
			return new self( (int) $post_id, (string) $taxonomy1, intval( $snapshot_id ?? '' ) );
		}
		return new self( intval( $post_tax_string ), '' );
	}

	/**
	 * Create instance from array with post_id, taxonomy, maybe snapshot_id from SQL query result
	 *
	 * @throws Ahrefs_Seo_Exception When array does not have post_id index.
	 *
	 * @param array $post_tax_array
	 * @return Post_Tax
	 */
	public static function create_from_array( array $post_tax_array ) : self {
		if ( ! isset( $post_tax_array['post_id'] ) ) { // really post_id index always exists, because we create it from query results.
			throw new Ahrefs_Seo_Exception( sprintf( 'Initialize Post_Tax with empty post_id %s', wp_json_encode( $post_tax_array ) ) );
		}
		$post_id     = isset( $post_tax_array['post_id'] ) ? (int) $post_tax_array['post_id'] : 0;
		$taxonomy    = $post_tax_array['taxonomy'] ?? '';
		$snapshot_id = isset( $post_tax_array['snapshot_id'] ) ? (int) $post_tax_array['snapshot_id'] : null;

		return new self( (int) $post_id, (string) $taxonomy, $snapshot_id );
	}

	/**
	 * Load from array of post tax strings.
	 *
	 * @param string[] $post_tax_strings
	 * @return Post_Tax[]
	 */
	public static function create_from_strings( array $post_tax_strings ) : array {
		return array_map(
			function( string $post_tax_string ) {
				return Post_Tax::create_from_string( "$post_tax_string" );
			},
			$post_tax_strings
		);
	}

	/**
	 * Create id: same as post tax string
	 *
	 * @param int    $post_id
	 * @param string $taxonomy
	 * @param int    $snapshot_id
	 * @return string
	 */
	public static function get_post_tax_string( int $post_id, string $taxonomy, int $snapshot_id ) : string {
		return "$post_id" . self::SEPARATOR . $taxonomy . self::SEPARATOR . "$snapshot_id";
	}

	/**
	 * Convert post tax or post tax list to id/ids list
	 *
	 * @param Post_Tax|Post_Tax[] $post_tax_or_post_taxes
	 * @return string|string[]
	 */
	public static function id( $post_tax_or_post_taxes ) {
		return is_array( $post_tax_or_post_taxes ) ? array_map(
			function( $post_tax ) {
				return (string) $post_tax;
			},
			$post_tax_or_post_taxes
		) : (string) $post_tax_or_post_taxes;
	}
}
